using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public class LineGenerator : MonoBehaviour {
	
	public bool loop = false;
	public bool segmentate = false;
	public float maxlinelength = 10;
	public float extrudeamount = 1;
	public static int particledensity=5;
	
	public MeshFilter foreground;
	
	public LevelVertex[] vertices;
	
	void OnDrawGizmos() {
		if (maxlinelength<1)
		{
			return;
		}
		List<Vector3> points = CalcPoints(false,true);
		
		
		Vector3 e = new Vector3(0,0,2*extrudeamount);
		
		for (int i=0;i<points.Count-1;i++)
		{
			
			Gizmos.color = new Color(1,1,1,1);
			
			
			Gizmos.DrawLine(points[i]-e,points[i+1]-e);
			Gizmos.DrawLine(points[i]-e,points[i]+e);
			Gizmos.DrawLine(points[i]+e,points[i+1]+e);
			
			Gizmos.color = new Color(1,1,0,1);
			Gizmos.DrawRay(
				points[i]/2+points[i+1]/2,
				Vector3.Cross(points[i+1]-points[i],new Vector3(0,0,extrudeamount)).normalized
				);
			
			if (segmentate)
			{
				i++;
			}
			
		}
		
		Gizmos.color = new Color(0,0,0,1);
		for (int i=0;i<points.Count;i++)
		{
			Gizmos.DrawSphere(points[i],0.05f);
		}

	}
	
	// Use this for initialization
	void Start () {
		GenerateMeshes();
		
		if (foreground!=null)
		{
			foreground.mesh=GenerateForegroundMesh();
			foreground.mesh.RecalculateBounds();
		}
		
		for (int i=0;i<this.vertices.Length;i++)
		{
			this.vertices[i].gameObject.active=false;
		}
		
		ParticleEmitter pe = this.GetComponent<ParticleEmitter>();
		if (pe!=null)
		{
			List<Vector3> points = CalcPoints(true,true);
			pe.maxEmission=points.Count*particledensity;
			pe.minEmission=pe.maxEmission;
		}
	}
			
	private List<Vector3> CalcPoints(bool local, bool interpolate)
	{
		List<Vector3> points = new List<Vector3>();
		
		if (vertices==null)
			return points;
		
		if (vertices.Length==0)
			return points;						
		
		for (int i=0;i<vertices.Length;i++)
		{			
			if (vertices[i]==null)
			{
				ClearFree();
			}
			
			if (local)
				points.Add(vertices[i].transform.localPosition);
			else
				points.Add(vertices[i].transform.position);
		}
		
		
		if (loop)
		{
			if (local)
				points.Add(vertices[0].transform.localPosition);
			else
				points.Add(vertices[0].transform.position);
		}
		
		if (interpolate)
		{
			bool added=true;
			while (added)
			{
				added=false;
				for (int i=0;i<points.Count-1;i++)
				{
					if (Vector3.Distance(points[i],points[i+1])>maxlinelength)
					{
						points.Insert(i+1,(points[i]+points[i+1])/2);
						added=true;
					}
				}
			}
		}
		
		return points;
	}
	
	public void ClearFree()
	{
		List<LevelVertex> validvertices = new List<LevelVertex>();
		
		for (int i=0;i<vertices.Length;i++)
		{
			if (vertices[i]!=null)
			{
				validvertices.Add(vertices[i]);
			}
		}
			
		vertices = validvertices.ToArray();		
	}
	
	
	private Mesh GenerateForegroundMesh()
	{
		if (this.loop==false)
			return null;
		
		List<Vector3> points = CalcPoints(true,true);		
		points.RemoveAt(points.Count-1);
		
		List<int> pointsunaccounted = new List<int>();
		for (int i=0;i<points.Count;i++)
		{
			pointsunaccounted.Add(i);
		}
		
			
		List<int> triangles = new List<int>();
		
		int t=0;
		while (triangles.Count<3*(points.Count-2))
		{		
			t++;
			if (t>100*points.Count)
			{
				//Debug.Log("ERROR OUT");
				break;
			}
			
			//Debug.Log("remainingpoints:"+pointsunaccounted.Count);

			//go through the unaccounted points array
			for (int i=0;i<pointsunaccounted.Count;i++)
			{
				int[] triangle = {
						pointsunaccounted[i],
						pointsunaccounted[(i+1) % pointsunaccounted.Count],
						pointsunaccounted[(i+2) % pointsunaccounted.Count]
				};
				
				//Debug.Log("TESTING TRIANGLE : "+ triangle[0].ToString() + ", " + triangle[1].ToString() + ", " + triangle[2].ToString());
				if (!PositiveTri(points[triangle[0]],points[triangle[1]],points[triangle[2]]))
				{
					continue;
				}
				//Debug.Log("positive");
				
				//check no other points in this triangle
				bool ok=true;
				for (int j=0;j<pointsunaccounted.Count-3;j++)
				{
					int targetpoint = (i+3+j) % pointsunaccounted.Count;
					if (PointInTri(
							points[triangle[0]],
							points[triangle[1]],
							points[triangle[2]],
							points[pointsunaccounted[targetpoint]]))
						{
							ok=false;
							break;
						}
				}
						
				if (!ok)
				{
						//Debug.Log("reject");
						continue;
				}
				else
				{
						//Debug.Log("Adding triangle");
						//add new triangle
						triangles.Add(triangle[0]);
						triangles.Add(triangle[1]);
						triangles.Add(triangle[2]);
						
						//ear snip!
						pointsunaccounted.RemoveAt(i+1 % pointsunaccounted.Count);
						break;
				}
						
			}
		}
		
		string s = "";
		for (int i=0;i<triangles.Count;i++)
		{
			s+=triangles[i].ToString()+", ";
		}
		//Debug.Log(s);
		
		//now actually make the mesh
		Vector3[] vertices_proper = new Vector3[points.Count];
		for (int i=0;i<points.Count;i++)
		{
			vertices_proper[i]=points[i];
		}
		
		Vector2[] uvs_proper = new Vector2[points.Count];
		for (int i=0;i<points.Count;i++)
		{
			uvs_proper[i]=new Vector2(i/2==0?0:1,i%2==0?0:1);
		}
				
		int[] tris_proper = new int[triangles.Count];

		for (int i=0;i<triangles.Count;i++)
		{
			tris_proper[i]=triangles[i];
		}
		
		Mesh mesh = new Mesh();
		
		mesh.Clear();
		mesh.vertices=vertices_proper;
		mesh.uv = uvs_proper;
		mesh.triangles=tris_proper;		
		mesh.RecalculateNormals();
		mesh.Optimize();
		mesh.RecalculateBounds();
		return mesh;
	}
	
	private bool PositiveTri(Vector3 a, Vector3 b, Vector3 c) {
		Vector3 d1 = b-a;
		Vector3 d2 = c-a;
		return Vector3.Cross(d1,d2).z<=0;
	}
	
	 /* Returns true if point p lies inside triangle a-b-c - assumes z coord is zero*/
	private bool PointInTri(Vector3 a, Vector3 b, Vector3 c, Vector3 p)  {
	    Vector3 v0 = c-a;
	    Vector3 v1 = b-a;
	    Vector3 v2 = p-a;
	    float dot00 = Vector3.Dot(v0,v0);
	    float dot01 = Vector3.Dot(v0,v1);
	    float dot02 = Vector3.Dot(v0,v2);
	    float dot11 = Vector3.Dot(v1,v1);
	    float dot12 = Vector3.Dot(v1,v2);
	    float invDenom = 1.0f / (dot00 * dot11 - dot01 * dot01);
	    float u = (dot11 * dot02 - dot01 * dot12) * invDenom;
	    float v = (dot00 * dot12 - dot01 * dot02) * invDenom;
	    return (u > 0.0f) && (v > 0.0f) && (u + v < 1.0f);
	}
		
	
	private Mesh GenerateMesh(float eamount, bool interpolate)
	{
		// don't interpolate when segmentate is on
		List<Vector3> points = CalcPoints(true,interpolate&&!segmentate);
		
		Vector3 e = new Vector3(0,0,2*eamount);	
		
		//now actually make the mesh
		Vector3[] vertices_proper = new Vector3[points.Count*2];
		for (int i=0;i<points.Count;i++)
		{
			vertices_proper[2*i]=points[i]-e;
			vertices_proper[2*i+1]=points[i]+e;
		}
		
		Vector2[] uvs_proper = new Vector2[points.Count*2];
		for (int i=0;i<points.Count;i++)
		{
			uvs_proper[2*i]=new Vector2(0,i%2==0 ? 0 : 1);
			uvs_proper[2*i+1]=new Vector2(1,i%2==0 ? 0 : 1);
		}
		
		
		int[] tris_proper = new int[Mathf.Max(0,(points.Count-1)*2*3)];

		for (int i=0;i<(points.Count-1);i++)
		{
			tris_proper[6*i]=2*i;
			tris_proper[6*i+1]=2*i+1;
			tris_proper[6*i+2]=2*i+2;
			tris_proper[6*i+3]=2*i+1;
			tris_proper[6*i+4]=2*i+2;
			tris_proper[6*i+5]=2*i+3;
			
			if (segmentate)
			{
				i++;
			}
		}
		
		Mesh mesh = new Mesh();
		
		mesh.Clear();
		mesh.vertices=vertices_proper;
		mesh.uv = uvs_proper;
		mesh.triangles=tris_proper;		
		mesh.RecalculateNormals();
		mesh.Optimize();
		return mesh;
		
	}
	
	public void GenerateMeshes()
	{
		
		this.GetComponent<MeshFilter>().mesh=GenerateMesh(extrudeamount,true);
		this.GetComponent<MeshFilter>().mesh.RecalculateBounds();
		
		MeshCollider mc = this.GetComponent<MeshCollider>();
		if (mc!=null)
		{
			mc.sharedMesh=GenerateMesh(extrudeamount,false);
			mc.sharedMesh.RecalculateBounds();
		}
	}
}
